@@ -19,6 +19,9 @@ class MockLogging(object):
         print ()
 
 #log = MockLogging()
+use_expanded_output = False
+def is_expanded_output():
+    return use_expanded_output
 
 def parse_special_command(sql):
     command, _, arg = sql.partition(' ')
@@ -27,6 +30,32 @@ def parse_special_command(sql):
     command = command.strip().replace('+', '')
     return (command, verbose, arg.strip())
 
+def list_schemas(cur, pattern, verbose):
+    """
+    Returns (rows, headers, status)
+    """
+
+    sql = '''SELECT n.nspname AS "Name",
+    pg_catalog.pg_get_userbyid(n.nspowner) AS "Owner"''' + (''',
+    pg_catalog.array_to_string(n.nspacl, E'\\n') AS "Access privileges",
+    pg_catalog.obj_description(n.oid, 'pg_namespace') AS "Description"''' if verbose else '') + """
+    FROM pg_catalog.pg_namespace n WHERE n.nspname """
+
+    params = []
+    if pattern:
+        _, schema = sql_name_pattern(pattern)
+        sql += '~ %s'
+        params.append(schema)
+    else:
+        sql += "!~ '^pg_' AND n.nspname <> 'information_schema'"
+    sql = cur.mogrify(sql + " ORDER BY 1", params)
+
+    log.debug(sql)
+    cur.execute(sql)
+    if cur.description:
+        headers = [x[0] for x in cur.description]
+        return [(cur.fetchall(), headers, cur.statusmessage)]
+
 def describe_table_details(cur, pattern, verbose):
     """
     Returns (rows, headers, status)
@@ -34,22 +63,22 @@ def describe_table_details(cur, pattern, verbose):
 
     # This is a simple \d command. No table name to follow.
     if not pattern:
-        sql = """SELECT n.nspname as "Schema", c.relname as "Name", 
-                    CASE c.relkind WHEN 'r' THEN 'table' 
-                        WHEN 'v' THEN 'view' 
-                        WHEN 'm' THEN 'materialized view' 
-                        WHEN 'i' THEN 'index' 
-                        WHEN 'S' THEN 'sequence' 
-                        WHEN 's' THEN 'special' 
-                        WHEN 'f' THEN 'foreign table' 
+        sql = """SELECT n.nspname as "Schema", c.relname as "Name",
+                    CASE c.relkind WHEN 'r' THEN 'table'
+                        WHEN 'v' THEN 'view'
+                        WHEN 'm' THEN 'materialized view'
+                        WHEN 'i' THEN 'index'
+                        WHEN 'S' THEN 'sequence'
+                        WHEN 's' THEN 'special'
+                        WHEN 'f' THEN 'foreign table'
                     END as "Type",
-                    pg_catalog.pg_get_userbyid(c.relowner) as "Owner" 
-                FROM pg_catalog.pg_class c 
-                LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace 
-                WHERE c.relkind IN ('r','v','m','S','f','') 
-                AND n.nspname <> 'pg_catalog' 
-                AND n.nspname <> 'information_schema' 
-                AND n.nspname !~ '^pg_toast' 
+                    pg_catalog.pg_get_userbyid(c.relowner) as "Owner"
+                FROM pg_catalog.pg_class c
+                LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace
+                WHERE c.relkind IN ('r','v','m','S','f','')
+                AND n.nspname <> 'pg_catalog'
+                AND n.nspname <> 'information_schema'
+                AND n.nspname !~ '^pg_toast'
                 AND pg_catalog.pg_table_is_visible(c.oid)
                 ORDER BY 1,2 """
 
@@ -61,15 +90,26 @@ def describe_table_details(cur, pattern, verbose):
 
     # This is a \d <tablename> command. A royal pain in the ass.
     schema, relname = sql_name_pattern(pattern)
-    sql ="""SELECT c.oid, n.nspname, c.relname 
-            FROM pg_catalog.pg_class c 
-            LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace 
-                AND pg_catalog.pg_table_is_visible(c.oid)
-            %s
-            %s
-            ORDER BY 2,3
-        """ % ( ' WHERE c.relname ~ ' + relname if relname else '',
-                ' AND n.nspname ~ ' + schema if schema else '' )
+    where = []
+    params = []
+
+    if not pattern:
+        where.append('pg_catalog.pg_table_is_visible(c.oid)')
+
+    if schema:
+        where.append('n.nspname ~ %s')
+        params.append(schema)
+
+    if relname:
+        where.append('c.relname ~ %s')
+        params.append(relname)
+
+    sql = """SELECT c.oid, n.nspname, c.relname
+             FROM pg_catalog.pg_class c
+             LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace
+             """ + ('WHERE ' + ' AND '.join(where) if where else '') + """
+             ORDER BY 2,3"""
+    sql = cur.mogrify(sql, params)
 
     # Execute the sql, get the results and call describe_one_table_details on each table.
 
@@ -91,16 +131,16 @@ def describe_one_table_details(cur, schema_name, relation_name, oid, verbose):
     else:
         suffix = "''"
 
-    sql ="""SELECT c.relchecks, c.relkind, c.relhasindex, 
-                c.relhasrules, c.relhastriggers, c.relhasoids, 
-                %s, 
-                c.reltablespace, 
-                CASE WHEN c.reloftype = 0 THEN '' 
-                    ELSE c.reloftype::pg_catalog.regtype::pg_catalog.text 
+    sql ="""SELECT c.relchecks, c.relkind, c.relhasindex,
+                c.relhasrules, c.relhastriggers, c.relhasoids,
+                %s,
+                c.reltablespace,
+                CASE WHEN c.reloftype = 0 THEN ''
+                    ELSE c.reloftype::pg_catalog.regtype::pg_catalog.text
                 END,
-                c.relpersistence 
-            FROM pg_catalog.pg_class c 
-            LEFT JOIN pg_catalog.pg_class tc ON (c.reltoastrelid = tc.oid) 
+                c.relpersistence
+            FROM pg_catalog.pg_class c
+            LEFT JOIN pg_catalog.pg_class tc ON (c.reltoastrelid = tc.oid)
             WHERE c.oid = '%s'""" % (suffix, oid)
 
     # Create a namedtuple called tableinfo and match what's in describe.c
@@ -675,19 +715,47 @@ def sql_name_pattern(pattern):
     used in a WHERE clause.
 
     Returns: schema_pattern, table_pattern
+
+    >>> sql_name_pattern('foo*."b""$ar*"')
+    ('^(foo.*)$', '^(b"\\\\$ar\\\\*)$')
     """
 
-    def replacements(pattern):
-        result = pattern.replace('*', '.*')
-        result = result.replace('?', '.')
-        result = result.replace('$', '\\$')
-        return result
+    inquotes = False
+    relname = ''
+    schema = None
+    pattern_len = len(pattern)
+    i = 0
+
+    while i < pattern_len:
+        c = pattern[i]
+        if c == '"':
+            if inquotes and i + 1 < pattern_len and pattern[i + 1] == '"':
+                relname += '"'
+                i += 1
+            else:
+                inquotes = not inquotes
+        elif not inquotes and c.isupper():
+            relname += c.lower()
+        elif not inquotes and c == '*':
+            relname += '.*'
+        elif not inquotes and c == '?':
+            relname += '.'
+        elif not inquotes and c == '.':
+            # Found schema/name separator, move current pattern to schema
+            schema = relname
+            relname = ''
+        else:
+            # Dollar is always quoted, whether inside quotes or not.
+            if c == '$' or inquotes and c in '|*+?()[]{}.^\\':
+                relname += '\\'
+            relname += c
+        i += 1
 
-    schema, _, relname = pattern.rpartition('.')
-    if schema:
-        schema = "'^(" + replacements(schema) + ")$'"
     if relname:
-        relname = "'^(" + replacements(relname) + ")$'"
+        relname = '^(' + relname + ')$'
+
+    if schema:
+        schema = '^(' + schema + ')$'
 
     return schema, relname
 
@@ -702,12 +770,20 @@ def show_help(cur, arg, verbose):  # All the parameters are ignored.
 def change_db(cur, arg, verbose):
     raise NotImplementedError
 
+def expanded_output(cur, arg, verbose):
+    global use_expanded_output
+    use_expanded_output = not use_expanded_output
+    message = u"Expanded display is "
+    message += u"on" if use_expanded_output else u"off"
+    return [(None, None, message + u".")]
 
 CASE_SENSITIVE_COMMANDS = {
             '\?': (show_help, ['\?', 'Help on pgcli commands.']),
             '\c': (change_db, ['\c database_name', 'Connect to a new database.']),
             '\l': ('''SELECT datname FROM pg_database;''', ['\l', 'list databases.']),
             '\d': (describe_table_details, ['\d [pattern]', 'list or describe tables, views and sequences.']),
+            '\dn': (list_schemas, ['\dn[+] [pattern]', 'list schemas']),
+            '\\x': (expanded_output, ['\\x', 'Toggle expanded output.']),
             '\dt': ('''SELECT n.nspname as "Schema", c.relname as "Name", CASE
             c.relkind WHEN 'r' THEN 'table' WHEN 'v' THEN 'view' WHEN 'm' THEN
             'materialized view' WHEN 'i' THEN 'index' WHEN 'S' THEN 'sequence'
