@@ -1,6 +1,8 @@
 import logging
 import psycopg2
+import psycopg2.extras
 import psycopg2.extensions
+from collections import defaultdict
 from .packages import pgspecial
 
 _logger = logging.getLogger(__name__)
@@ -10,6 +12,10 @@ _logger = logging.getLogger(__name__)
 psycopg2.extensions.register_type(psycopg2.extensions.UNICODE)
 psycopg2.extensions.register_type(psycopg2.extensions.UNICODEARRAY)
 
+# When running a query, make pressing CTRL+C raise a KeyboardInterrupt
+# See http://initd.org/psycopg/articles/2014/07/20/cancelling-postgresql-statements-python/
+psycopg2.extensions.set_wait_callback(psycopg2.extras.wait_select)
+
 def _parse_dsn(dsn, default_user, default_password, default_host,
         default_port):
     """
@@ -53,8 +59,7 @@ class PGExecute(object):
     'information_schema' AND n.nspname !~ '^pg_toast' AND
     pg_catalog.pg_table_is_visible(c.oid) ORDER BY 1;'''
 
-    columns_query = '''SELECT column_name FROM information_schema.columns WHERE
-    table_name =%s;'''
+    columns_query = '''SELECT table_name, column_name FROM information_schema.columns'''
 
     databases_query = """SELECT d.datname as "Name",
        pg_catalog.pg_get_userbyid(d.datdba) as "Owner",
@@ -70,6 +75,11 @@ class PGExecute(object):
                 _parse_dsn(database, default_user=user,
                         default_password=password, default_host=host,
                         default_port=port)
+        self.connect()
+
+    def connect(self):
+        if hasattr(self, 'conn'):
+            self.conn.close()
         self.conn = psycopg2.connect(database=self.dbname, user=self.user,
                 password=self.password, host=self.host, port=self.port)
         self.conn.autocommit = True
@@ -123,21 +133,19 @@ class PGExecute(object):
                 return [(None, None, cur.statusmessage)]
 
     def tables(self):
+        """ Returns tuple (sorted_tables, columns). Columns is a dictionary of
+            table name -> list of columns """
+        columns = defaultdict(list)
         with self.conn.cursor() as cur:
             cur.execute(self.tables_query)
-            return [x[0] for x in cur.fetchall()]
+            tables = [x[0] for x in cur.fetchall()]
 
-    def columns(self, table):
-        with self.conn.cursor() as cur:
-            cur.execute(self.columns_query, (table,))
-            cols = [x[0] for x in cur.fetchall()]
-            return cols
-
-    def all_columns(self):
-        columns = set()
-        for table in self.tables():
-            columns.update(self.columns(table))
-        return columns
+            table_set = set(tables)
+            cur.execute(self.columns_query)
+            for table, column in cur.fetchall():
+                if table in table_set:
+                    columns[table].append(column)
+        return tables, columns
 
     def databases(self):
         with self.conn.cursor() as cur:
